//     fslmerge.cc concatenate AVW files into a single output
//     Steve Smith, David Flitney, Stuart Clare and Matthew Webster, FMRIB Image Analysis Group
//     Copyright (C) 2000-2019 University of Oxford
//     CCOPYRIGHT
#define EXPOSE_TREACHEROUS
#include "newimage/newimageall.h"
#include "newimage/fmribmain.h"
using namespace NEWIMAGE;

void print_usage(const string& progname)
{
  cout << endl;
  cout << "Usage: fslmerge <-x/y/z/t/a/tr> <output> <file1 file2 .......> [tr value in seconds]" << endl;
  cout << "     -t : concatenate images in time" << endl;
  cout << "     -x : concatenate images in the x direction"  << endl;
  cout << "     -y : concatenate images in the y direction"  << endl;
  cout << "     -z : concatenate images in the z direction" << endl;
  cout << "     -a : auto-choose: single slices -> volume, volumes -> 4D (time series)"  << endl;
  cout << "     -tr : concatenate images in time and set the output image tr to the final option value"  << endl;
}

template <class T>
int fmrib_main(int argc, char *argv[])
{
  volume4D<T> input_volume;
  int direction,dimerror=0,xoffset=0,yoffset=0,zoffset=0,toffset=0;

  NiftiHeader header(loadHeader(return_validimagefilename(string(argv[3]))));
  float newTR(header.pixdim[4]);

  if (!strcmp(argv[1], "-t"))       direction=0;
  else if (!strcmp(argv[1], "-x"))  direction=1;
  else if (!strcmp(argv[1], "-y"))  direction=2;
  else if (!strcmp(argv[1], "-z"))  direction=3;
  else if (!strcmp(argv[1], "-a"))  direction=4;
  else if (!strcmp(argv[1], "-tr")) {
    direction=0;
    newTR=atof(argv[--argc]);
  }
  else
  {
    print_usage(string(argv[0]));
    return(1);
  }

  int xdimtot(std::max(header.dim[1],(int64_t)1)), ydimtot(std::max(header.dim[2],(int64_t)1)), zdimtot(std::max(header.dim[3],(int64_t)1)), tdimtot(std::max(header.dim[4],(int64_t)1));

  if(direction==4)
  {
    if( zdimtot<2 && tdimtot<2 ) direction=3;
     else direction=0;
  }

  for(int vol = 4; vol < argc; vol++)
  {
    header=loadHeader(return_validimagefilename(string(argv[vol])));
    if (direction==0) tdimtot+=std::max(header.dim[4],(int64_t)1);
    if (direction==1) xdimtot+=std::max(header.dim[1],(int64_t)1);
    if (direction==2) ydimtot+=std::max(header.dim[2],(int64_t)1);
    if (direction==3) zdimtot+=std::max(header.dim[3],(int64_t)1);
  }
  volume4D<T> output_volume(xdimtot,ydimtot,zdimtot,tdimtot);
  read_volume4D(input_volume,string(argv[3]));
  output_volume.copyproperties(input_volume);

  for(int vol = 3; vol < argc; vol++)
  {
    if (vol>3) {
      read_volume4D(input_volume,string(argv[vol]));
      // sanity check on valid orientation info (it should be consistent)
      if ((output_volume.sform_code()!=NIFTI_XFORM_UNKNOWN) || (output_volume.qform_code()!=NIFTI_XFORM_UNKNOWN)) {
	if ((input_volume.sform_code()!=NIFTI_XFORM_UNKNOWN) || (input_volume.qform_code()!=NIFTI_XFORM_UNKNOWN)) {
	  float rms = rms_deviation(output_volume.newimagevox2mm_mat(),input_volume.newimagevox2mm_mat());
	  if ( rms > 0.5 && direction == 0 ) {   // arbitrary 0.5mm rms diff threshold - maybe too sensitive?
	    cerr << endl << "WARNING:: Inconsistent orientations for individual images when attempting to merge." <<endl;
	    cerr <<"          Merge will use voxel-based orientation which is probably incorrect - *PLEASE CHECK*!" <<endl<<endl;
	  }
	}
      }
    }
    if (direction == 0 && (input_volume.xsize() != xdimtot || input_volume.ysize() != ydimtot || input_volume.zsize() != zdimtot)) dimerror=1;
    if (direction == 1 && (input_volume.ysize() != ydimtot || input_volume.zsize() != zdimtot || input_volume.tsize() != tdimtot)) dimerror=1;
    if (direction == 2 && (input_volume.xsize() != xdimtot || input_volume.zsize() != zdimtot || input_volume.tsize() != tdimtot)) dimerror=1;
    if (direction == 3 && (input_volume.xsize() != xdimtot || input_volume.ysize() != ydimtot || input_volume.tsize() != tdimtot)) dimerror=1;
    if (dimerror)
    {
      cerr << "Error in size-match along non-concatenated dimension for input file: " << string(argv[vol]) << endl;
      return 1;
    }


    for(int t=0;t<input_volume.tsize();t++)
      for(int k=0;k<input_volume.zsize();k++)
        for(int j=0;j<input_volume.ysize();j++)
          for(int i=0;i<input_volume.xsize();i++)
            output_volume.value(i+xoffset,j+yoffset,k+zoffset,t+toffset)=input_volume.value(i,j,k,t);
    if (direction==0)  toffset+=input_volume.tsize();
    if (direction==1)  xoffset+=input_volume.xsize();
    if (direction==2)  yoffset+=input_volume.ysize();
    if (direction==3)  zoffset+=input_volume.zsize();
  }
  output_volume.setTR(newTR);
  save_volume4D(output_volume,string(argv[2]));
  return 0;
}


int main(int argc,char *argv[])
{
  if (argc < 4)
  {
    print_usage(string(argv[0]));
    return 1;
  }
  return call_fmrib_main(dtype(string(argv[3])),argc,argv);
}
